@@ -14,6 +14,7 @@
 
 import base64
 import functools
+import logging
 import threading
 import uuid
 
@@ -27,6 +28,8 @@ from tooz import coordination
 from tooz import locking
 from tooz import utils
 
+LOG = logging.getLogger(__name__)
+
 
 def _encode(data):
     """Safely encode data for consumption of the gateway."""
@@ -177,6 +180,7 @@ class Etcd3Driver(coordination.CoordinationDriverWithExecutor):
     ==================  =======
     Name                Default
     ==================  =======
+    api_version         v3alpha
     ca_cert             None
     cert_key            None
     cert_cert           None
@@ -195,6 +199,9 @@ class Etcd3Driver(coordination.CoordinationDriverWithExecutor):
     #: Default port used if none provided (4001 or 2379 are the common ones).
     DEFAULT_PORT = 2379
 
+    #: Default api version if none provided
+    DEFAULT_API_VERSION = "v3alpha"
+
     GROUP_PREFIX = b"tooz/groups/"
 
     def __init__(self, member_id, parsed_url, options):
@@ -207,17 +214,20 @@ class Etcd3Driver(coordination.CoordinationDriverWithExecutor):
         cert_key = options.get('cert_key')
         cert_cert = options.get('cert_cert')
         timeout = int(options.get('timeout', self.DEFAULT_TIMEOUT))
+        api_version = options.get("api_version", self.DEFAULT_API_VERSION)
         self.client = etcd3gw.client(host=host,
                                      port=port,
                                      protocol=protocol,
                                      ca_cert=ca_cert,
                                      cert_key=cert_key,
                                      cert_cert=cert_cert,
+                                     api_path="/" + api_version + "/",
                                      timeout=timeout)
         self.lock_timeout = int(options.get('lock_timeout', timeout))
         self.membership_timeout = int(options.get(
             'membership_timeout', timeout))
         self._acquired_locks = set()
+        self._membership_lease = None
 
     def _start(self):
         super(Etcd3Driver, self)._start()
@@ -227,11 +237,15 @@ class Etcd3Driver(coordination.CoordinationDriverWithExecutor):
         return Etcd3Lock(self, name, self.lock_timeout)
 
     def heartbeat(self):
+        # TODO(kaifeng) use the same lease for locks?
+        if self._membership_lease.refresh() == -1:
+            expired_lease = self._membership_lease.id
+            self._membership_lease = self.client.lease(self.membership_timeout)
+            LOG.debug('Created new lease %s after previous lease %s expired.',
+                      self._membership_lease.id, expired_lease)
         # NOTE(jaypipes): Copying because set can mutate during iteration
         for lock in self._acquired_locks.copy():
             lock.heartbeat()
-        # TODO(kaifeng) use the same lease for locks?
-        self._membership_lease.refresh()
         return min(self.lock_timeout, self.membership_timeout)
 
     def watch_join_group(self, group_id, callback):
@@ -250,7 +264,7 @@ class Etcd3Driver(coordination.CoordinationDriverWithExecutor):
         return _encode(self._prefix_group(group_id))
 
     def _prefix_group(self, group_id):
-        return b"%s%s/" % (self.GROUP_PREFIX, group_id)
+        return b"%s%s/" % (self.GROUP_PREFIX, utils.to_binary(group_id))
 
     def create_group(self, group_id):
         @_translate_failures
