@@ -14,18 +14,18 @@
 # License for the specific language governing permissions and limitations
 # under the License.
 
-import collections
+import contextlib
 import errno
+import functools
 import logging
 import socket
 
-from concurrent import futures
 from oslo_utils import encodeutils
 from pymemcache import client as pymemcache_client
-import six
 
+import tooz
+from tooz import _retry
 from tooz import coordination
-from tooz.drivers import _retry
 from tooz import locking
 from tooz import utils
 
@@ -33,35 +33,41 @@ from tooz import utils
 LOG = logging.getLogger(__name__)
 
 
-def _translate_failures(func):
+@contextlib.contextmanager
+def _failure_translator():
     """Translates common pymemcache exceptions into tooz exceptions.
 
     https://github.com/pinterest/pymemcache/blob/d995/pymemcache/client.py#L202
     """
+    try:
+        yield
+    except pymemcache_client.MemcacheUnexpectedCloseError as e:
+        utils.raise_with_cause(coordination.ToozConnectionError,
+                               encodeutils.exception_to_unicode(e),
+                               cause=e)
+    except (socket.timeout, socket.error,
+            socket.gaierror, socket.herror) as e:
+        # TODO(harlowja): get upstream pymemcache to produce a better
+        # exception for these, using socket (vs. a memcache specific
+        # error) seems sorta not right and/or the best approach...
+        msg = encodeutils.exception_to_unicode(e)
+        if e.errno is not None:
+            msg += " (with errno %s [%s])" % (errno.errorcode[e.errno],
+                                              e.errno)
+        utils.raise_with_cause(coordination.ToozConnectionError,
+                               msg, cause=e)
+    except pymemcache_client.MemcacheError as e:
+        utils.raise_with_cause(tooz.ToozError,
+                               encodeutils.exception_to_unicode(e),
+                               cause=e)
 
-    @six.wraps(func)
+
+def _translate_failures(func):
+
+    @functools.wraps(func)
     def wrapper(*args, **kwargs):
-        try:
+        with _failure_translator():
             return func(*args, **kwargs)
-        except pymemcache_client.MemcacheUnexpectedCloseError as e:
-            coordination.raise_with_cause(coordination.ToozConnectionError,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
-        except (socket.timeout, socket.error,
-                socket.gaierror, socket.herror) as e:
-            # TODO(harlowja): get upstream pymemcache to produce a better
-            # exception for these, using socket (vs. a memcache specific
-            # error) seems sorta not right and/or the best approach...
-            msg = encodeutils.exception_to_unicode(e)
-            if e.errno is not None:
-                msg += " (with errno %s [%s])" % (errno.errorcode[e.errno],
-                                                  e.errno)
-            coordination.raise_with_cause(coordination.ToozConnectionError,
-                                          msg, cause=e)
-        except pymemcache_client.MemcacheError as e:
-            coordination.raise_with_cause(coordination.ToozError,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
 
     return wrapper
 
@@ -83,7 +89,9 @@ class MemcachedLock(locking.Lock):
                 return False
             return owner == self.coord._member_id
 
-    def acquire(self, blocking=True):
+    def acquire(self, blocking=True, shared=False):
+        if shared:
+            raise tooz.NotImplemented
 
         @_retry.retry(stop_max_delay=blocking)
         @_translate_failures
@@ -97,7 +105,7 @@ class MemcachedLock(locking.Lock):
                 return True
             if blocking is False:
                 return False
-            raise _retry.Retry
+            raise _retry.TryAgain
 
         return _acquire()
 
@@ -142,11 +150,16 @@ class MemcachedLock(locking.Lock):
         # it being done in the client side (non-atomic).
         value = self.coord.client.get(self.name)
         if value != self.coord._member_id:
+            # NOTE(zhen): Although ``member_ id`` is different, self lock
+            # object needs to be removed from'_ acquired_locks' because it
+            # has the same key.
+            self.coord._acquired_locks.remove(self)
             return False
         else:
+            # NOTE(zhen): Whether 'was_deleted' was 'TRUE' or not,
+            # eventually we have to remove self from '_acquired_locks'.
             was_deleted = self.coord.client.delete(self.name, noreply=False)
-            if was_deleted:
-                self.coord._acquired_locks.remove(self)
+            self.coord._acquired_locks.remove(self)
             return was_deleted
 
     @_translate_failures
@@ -156,10 +169,12 @@ class MemcachedLock(locking.Lock):
             poked = self.coord.client.touch(self.name,
                                             expire=self.timeout,
                                             noreply=False)
-            if not poked:
-                LOG.warn("Unable to heartbeat by updating key '%s' with"
-                         " extended expiry of %s seconds", self.name,
-                         self.timeout)
+            if poked:
+                return True
+            LOG.warning("Unable to heartbeat by updating key '%s' with "
+                        "extended expiry of %s seconds", self.name,
+                        self.timeout)
+        return False
 
     @_translate_failures
     def get_owner(self):
@@ -170,8 +185,8 @@ class MemcachedLock(locking.Lock):
         return self in self.coord._acquired_locks
 
 
-class MemcachedDriver(coordination._RunWatchersMixin,
-                      coordination.CoordinationDriver):
+class MemcachedDriver(coordination.CoordinationDriverCachedRunWatchers,
+                      coordination.CoordinationDriverWithExecutor):
     """A `memcached`_ based driver.
 
     This driver users `memcached`_ concepts to provide the coordination driver
@@ -179,6 +194,23 @@ class MemcachedDriver(coordination._RunWatchersMixin,
     all of the coordination driver API(s). It stores data into memcache
     using expiries and `msgpack`_ encoded values.
 
+    The Memcached driver connection URI should look like::
+
+      memcached://[HOST[:PORT]][?OPTION1=VALUE1[&OPTION2=VALUE2[&...]]]
+
+    If not specified, HOST defaults to localhost and PORT defaults to 11211.
+    Available options are:
+
+    ==================  =======
+    Name                Default
+    ==================  =======
+    timeout             30
+    membership_timeout  30
+    lock_timeout        30
+    leader_timeout      30
+    max_pool_size       None
+    ==================  =======
+
     General recommendations/usage considerations:
 
     - Memcache (without different backend technology) is a **cache** enough
@@ -218,23 +250,18 @@ class MemcachedDriver(coordination._RunWatchersMixin,
     STILL_ALIVE = b"It's alive!"
 
     def __init__(self, member_id, parsed_url, options):
-        super(MemcachedDriver, self).__init__()
-        options = utils.collapse(options)
-        self._options = options
-        self._member_id = member_id
-        self._joined_groups = set()
-        self._executor = utils.ProxyExecutor.build("Memcached", options)
+        super(MemcachedDriver, self).__init__(member_id, parsed_url, options)
         self.host = (parsed_url.hostname or "localhost",
                      parsed_url.port or 11211)
-        default_timeout = options.get('timeout', self.DEFAULT_TIMEOUT)
+        default_timeout = self._options.get('timeout', self.DEFAULT_TIMEOUT)
         self.timeout = int(default_timeout)
-        self.membership_timeout = int(options.get(
+        self.membership_timeout = int(self._options.get(
             'membership_timeout', default_timeout))
-        self.lock_timeout = int(options.get(
+        self.lock_timeout = int(self._options.get(
             'lock_timeout', default_timeout))
-        self.leader_timeout = int(options.get(
+        self.leader_timeout = int(self._options.get(
             'leader_timeout', default_timeout))
-        max_pool_size = options.get('max_pool_size', None)
+        max_pool_size = self._options.get('max_pool_size', None)
         if max_pool_size is not None:
             self.max_pool_size = int(max_pool_size)
         else:
@@ -243,7 +270,7 @@ class MemcachedDriver(coordination._RunWatchersMixin,
 
     @staticmethod
     def _msgpack_serializer(key, value):
-        if isinstance(value, six.binary_type):
+        if isinstance(value, bytes):
             return value, 1
         return utils.dumps(value), 2
 
@@ -258,6 +285,7 @@ class MemcachedDriver(coordination._RunWatchersMixin,
 
     @_translate_failures
     def _start(self):
+        super(MemcachedDriver, self)._start()
         self.client = pymemcache_client.PooledClient(
             self.host,
             serializer=self._msgpack_serializer,
@@ -268,34 +296,23 @@ class MemcachedDriver(coordination._RunWatchersMixin,
         # Run heartbeat here because pymemcache use a lazy connection
         # method and only connect once you do an operation.
         self.heartbeat()
-        self._group_members = collections.defaultdict(set)
-        self._executor.start()
 
     @_translate_failures
     def _stop(self):
+        super(MemcachedDriver, self)._stop()
         for lock in list(self._acquired_locks):
             lock.release()
         self.client.delete(self._encode_member_id(self._member_id))
-        for g in list(self._joined_groups):
-            try:
-                self.leave_group(g).get()
-            except (coordination.MemberNotJoined,
-                    coordination.GroupNotCreated):
-                # Guess we got booted out/never existed in the first place...
-                pass
-            except coordination.ToozError:
-                LOG.warning("Unable to leave group '%s'", g, exc_info=True)
-        self._executor.stop()
         self.client.close()
 
     def _encode_group_id(self, group_id):
-        return self.GROUP_PREFIX + group_id
+        return self.GROUP_PREFIX + utils.to_binary(group_id)
 
     def _encode_member_id(self, member_id):
-        return self.MEMBER_PREFIX + member_id
+        return self.MEMBER_PREFIX + utils.to_binary(member_id)
 
     def _encode_group_leader(self, group_id):
-        return self.GROUP_LEADER_PREFIX + group_id
+        return self.GROUP_LEADER_PREFIX + utils.to_binary(group_id)
 
     @_retry.retry()
     def _add_group_to_group_list(self, group_id):
@@ -310,12 +327,12 @@ class MemcachedDriver(coordination._RunWatchersMixin,
             if not self.client.cas(self.GROUP_LIST_KEY,
                                    list(group_list), cas):
                 # Someone updated the group list before us, try again!
-                raise _retry.Retry
+                raise _retry.TryAgain
         else:
             if not self.client.add(self.GROUP_LIST_KEY,
                                    [group_id], noreply=False):
                 # Someone updated the group list before us, try again!
-                raise _retry.Retry
+                raise _retry.TryAgain
 
     @_retry.retry()
     def _remove_from_group_list(self, group_id):
@@ -329,7 +346,7 @@ class MemcachedDriver(coordination._RunWatchersMixin,
         if not self.client.cas(self.GROUP_LIST_KEY,
                                list(group_list), cas):
             # Someone updated the group list before us, try again!
-            raise _retry.Retry
+            raise _retry.TryAgain
 
     def create_group(self, group_id):
         encoded_group = self._encode_group_id(group_id)
@@ -367,7 +384,7 @@ class MemcachedDriver(coordination._RunWatchersMixin,
             }
             if not self.client.cas(encoded_group, group_members, cas):
                 # It changed, let's try again
-                raise _retry.Retry
+                raise _retry.TryAgain
             self._joined_groups.add(group_id)
 
         return MemcachedFutureResult(self._executor.submit(_join_group))
@@ -386,7 +403,7 @@ class MemcachedDriver(coordination._RunWatchersMixin,
             del group_members[self._member_id]
             if not self.client.cas(encoded_group, group_members, cas):
                 # It changed, let's try again
-                raise _retry.Retry
+                raise _retry.TryAgain
             self._joined_groups.discard(group_id)
 
         return MemcachedFutureResult(self._executor.submit(_leave_group))
@@ -408,7 +425,7 @@ class MemcachedDriver(coordination._RunWatchersMixin,
             # Delete is not atomic, so we first set the group to
             # using CAS, and then we delete it, to avoid race conditions.
             if not self.client.cas(encoded_group, None, cas):
-                raise _retry.Retry
+                raise _retry.TryAgain
             self.client.delete(encoded_group)
             self._remove_from_group_list(group_id)
 
@@ -422,7 +439,7 @@ class MemcachedDriver(coordination._RunWatchersMixin,
         if group_members is None:
             raise coordination.GroupNotCreated(group_id)
         actual_group_members = {}
-        for m, v in six.iteritems(group_members):
+        for m, v in group_members.items():
             # Never kick self from the group, we know we're alive
             if (m == self._member_id or
                self.client.get(self._encode_member_id(m))):
@@ -431,13 +448,13 @@ class MemcachedDriver(coordination._RunWatchersMixin,
             # There are some dead members, update the group
             if not self.client.cas(encoded_group, actual_group_members, cas):
                 # It changed, let's try again
-                raise _retry.Retry
+                raise _retry.TryAgain
         return actual_group_members
 
     def get_members(self, group_id):
 
         def _get_members():
-            return self._get_members(group_id).keys()
+            return set(self._get_members(group_id).keys())
 
         return MemcachedFutureResult(self._executor.submit(_get_members))
 
@@ -466,7 +483,7 @@ class MemcachedDriver(coordination._RunWatchersMixin,
             group_members[self._member_id][b'capabilities'] = capabilities
             if not self.client.cas(encoded_group, group_members, cas):
                 # It changed, try again
-                raise _retry.Retry
+                raise _retry.TryAgain
 
         return MemcachedFutureResult(
             self._executor.submit(_update_capabilities))
@@ -486,41 +503,9 @@ class MemcachedDriver(coordination._RunWatchersMixin,
         # Reset the acquired locks
         for lock in self._acquired_locks:
             lock.heartbeat()
-
-    @_translate_failures
-    def _init_watch_group(self, group_id):
-        members = self.client.get(self._encode_group_id(group_id))
-        if members is None:
-            raise coordination.GroupNotCreated(group_id)
-        # Initialize with the current group member list
-        if group_id not in self._group_members:
-            self._group_members[group_id] = set(members.keys())
-
-    def watch_join_group(self, group_id, callback):
-        self._init_watch_group(group_id)
-        return super(MemcachedDriver, self).watch_join_group(
-            group_id, callback)
-
-    def unwatch_join_group(self, group_id, callback):
-        return super(MemcachedDriver, self).unwatch_join_group(
-            group_id, callback)
-
-    def watch_leave_group(self, group_id, callback):
-        self._init_watch_group(group_id)
-        return super(MemcachedDriver, self).watch_leave_group(
-            group_id, callback)
-
-    def unwatch_leave_group(self, group_id, callback):
-        return super(MemcachedDriver, self).unwatch_leave_group(
-            group_id, callback)
-
-    def watch_elected_as_leader(self, group_id, callback):
-        return super(MemcachedDriver, self).watch_elected_as_leader(
-            group_id, callback)
-
-    def unwatch_elected_as_leader(self, group_id, callback):
-        return super(MemcachedDriver, self).unwatch_elected_as_leader(
-            group_id, callback)
+        return min(self.membership_timeout,
+                   self.leader_timeout,
+                   self.lock_timeout)
 
     def get_lock(self, name):
         return MemcachedLock(self, name, self.lock_timeout)
@@ -531,7 +516,7 @@ class MemcachedDriver(coordination._RunWatchersMixin,
 
     @_translate_failures
     def run_elect_coordinator(self):
-        for group_id, hooks in six.iteritems(self._hooks_elected_leader):
+        for group_id, hooks in self._hooks_elected_leader.items():
             # Try to grab the lock, if that fails, that means someone has it
             # already.
             leader_lock = self._get_leader_lock(group_id)
@@ -547,19 +532,6 @@ class MemcachedDriver(coordination._RunWatchersMixin,
         return result
 
 
-class MemcachedFutureResult(coordination.CoordAsyncResult):
-    """Memcached asynchronous result that references a future."""
-    def __init__(self, fut):
-        self._fut = fut
-
-    def get(self, timeout=10):
-        try:
-            return self._fut.result(timeout=timeout)
-        except futures.TimeoutError as e:
-            coordination.raise_with_cause(
-                coordination.OperationTimedOut,
-                encodeutils.exception_to_unicode(e),
-                cause=e)
-
-    def done(self):
-        return self._fut.done()
+MemcachedFutureResult = functools.partial(
+    coordination.CoordinatorResult,
+    failure_translator=_failure_translator)
