@@ -61,12 +61,12 @@ class MySQLLock(locking.Lock):
             try:
                 if not self._conn.open:
                     self._conn.connect()
-                with self._conn as cur:
-                    cur.execute("SELECT GET_LOCK(%s, 0);", self.name)
-                    # Can return NULL on error
-                    if cur.fetchone()[0] is 1:
-                        self.acquired = True
-                        return True
+                cur = self._conn.cursor()
+                cur.execute("SELECT GET_LOCK(%s, 0);", self.name)
+                # Can return NULL on error
+                if cur.fetchone()[0] == 1:
+                    self.acquired = True
+                    return True
             except pymysql.MySQLError as e:
                 utils.raise_with_cause(
                     tooz.ToozError,
@@ -90,9 +90,9 @@ class MySQLLock(locking.Lock):
         if not self.acquired:
             return False
         try:
-            with self._conn as cur:
-                cur.execute("SELECT RELEASE_LOCK(%s);", self.name)
-                cur.fetchone()
+            cur = self._conn.cursor()
+            cur.execute("SELECT RELEASE_LOCK(%s);", self.name)
+            cur.fetchone()
             self.acquired = False
             self._conn.close()
             return True
@@ -114,6 +114,26 @@ class MySQLDriver(coordination.CoordinationDriver):
     missing some functionality but in the future these not implemented API(s)
     will be filled in.
 
+    The MySQL driver connection URI should look like::
+
+      mysql://USERNAME:PASSWORD@HOST[:PORT]/DBNAME[?OPTION1=VALUE1[&OPTION2=VALUE2[&...]]]
+
+    If not specified, PORT defaults to 3306.
+    Available options are:
+
+    ==================  =======
+    Name                Default
+    ==================  =======
+    ssl_ca              None
+    ssl_capath          None
+    ssl_cert            None
+    ssl_key             None
+    ssl_cipher          None
+    ssl_verify_mode     None
+    ssl_check_hostname  True
+    unix_socket         None
+    ==================  =======
+
     .. _MySQL: http://dev.mysql.com/
     """
 
@@ -176,6 +196,26 @@ class MySQLDriver(coordination.CoordinationDriver):
         username = parsed_url.username
         password = parsed_url.password
         unix_socket = options.get("unix_socket")
+        ssl_opt_names = (
+            "ca",
+            "capath",
+            "cert",
+            "key",
+            "cipher",
+            "verify_mode",
+        )
+        ssl_args = {}
+        for o in ssl_opt_names:
+            value = options.get("ssl_" + o)
+            if value:
+                ssl_args[o] = value
+        check_hostname = options.get("ssl_check_hostname")
+        if check_hostname is not None:
+            check_hostname = check_hostname.lower()
+            if check_hostname in ("true", "1", "yes"):
+                ssl_args["check_hostname"] = True
+            elif check_hostname in ("false", "0", "no"):
+                ssl_args["check_hostname"] = False
 
         try:
             if unix_socket:
@@ -184,6 +224,7 @@ class MySQLDriver(coordination.CoordinationDriver):
                                        user=username,
                                        passwd=password,
                                        database=dbname,
+                                       ssl=ssl_args,
                                        defer_connect=defer_connect)
             else:
                 return pymysql.Connect(host=host,
@@ -191,6 +232,7 @@ class MySQLDriver(coordination.CoordinationDriver):
                                        user=username,
                                        passwd=password,
                                        database=dbname,
+                                       ssl=ssl_args,
                                        defer_connect=defer_connect)
         except (pymysql.err.OperationalError, pymysql.err.InternalError) as e:
             utils.raise_with_cause(coordination.ToozConnectionError,
