@@ -14,19 +14,19 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-import collections
-import copy
-
 from kazoo import client
 from kazoo import exceptions
-from kazoo.handlers import eventlet as eventlet_handler
+from kazoo import security
+try:
+    from kazoo.handlers import eventlet as eventlet_handler
+except ImportError:
+    eventlet_handler = None
 from kazoo.handlers import threading as threading_handler
 from kazoo.protocol import paths
 from oslo_utils import encodeutils
 from oslo_utils import strutils
-import six
-from six.moves import filter as compat_filter
 
+import tooz
 from tooz import coordination
 from tooz import locking
 from tooz import utils
@@ -51,11 +51,13 @@ class ZooKeeperLock(locking.Lock):
                 exceptions.NoNodeError):
             return False
         except exceptions.KazooException as e:
-            coordination.raise_with_cause(coordination.ToozError,
-                                          "operation error: %s" % (e),
-                                          cause=e)
+            utils.raise_with_cause(tooz.ToozError,
+                                   "operation error: %s" % (e),
+                                   cause=e)
 
-    def acquire(self, blocking=True):
+    def acquire(self, blocking=True, shared=False):
+        if shared:
+            raise tooz.NotImplemented
         blocking, timeout = utils.convert_blocking(blocking)
         return self._lock.acquire(blocking=blocking,
                                   timeout=timeout)
@@ -72,40 +74,92 @@ class ZooKeeperLock(locking.Lock):
         return self._lock.is_acquired
 
 
-class BaseZooKeeperDriver(coordination.CoordinationDriver):
-    """Initialize the zookeeper driver.
+class KazooDriver(coordination.CoordinationDriverCachedRunWatchers):
+    """This driver uses the `kazoo`_ client against real `zookeeper`_ servers.
 
-    :param timeout: connection timeout to wait when first connecting to the
-                    zookeeper server
+    It **is** fully functional and implements all of the coordination
+    driver API(s). It stores data into `zookeeper`_ using znodes
+    and `msgpack`_ encoded values.
+
+    To configure the client to your liking a subset of the options defined at
+    http://kazoo.readthedocs.org/en/latest/api/client.html
+    will be extracted from the coordinator url (or any provided options),
+    so that a specific coordinator can be created that will work for you.
+
+    The Zookeeper coordinator url should look like::
+
+      zookeeper://[USERNAME:PASSWORD@][HOST[:PORT]][?OPTION1=VALUE1[&OPTION2=VALUE2[&...]]]
+
+    Currently the following options will be proxied to the contained client:
+
+    ================  ===============================  ====================
+    Name              Source                           Default
+    ================  ===============================  ====================
+    hosts             url netloc + 'hosts' option key  localhost:2181
+    timeout           'timeout' options key            10.0 (kazoo default)
+    connection_retry  'connection_retry' options key   None
+    command_retry     'command_retry' options key      None
+    randomize_hosts   'randomize_hosts' options key    True
+    ================  ===============================  ====================
+
+    .. _kazoo: http://kazoo.readthedocs.org/
+    .. _zookeeper: http://zookeeper.apache.org/
+    .. _msgpack: http://msgpack.org/
     """
     #: Default namespace when none is provided.
     TOOZ_NAMESPACE = b"tooz"
 
+    HANDLERS = {
+        'threading': threading_handler.SequentialThreadingHandler,
+    }
+
+    if eventlet_handler:
+        HANDLERS['eventlet'] = eventlet_handler.SequentialEventletHandler
+
+    """
+    Restricted immutable dict of handler 'kinds' -> handler classes that
+    this driver can accept via 'handler' option key (the expected value for
+    this option is one of the keys in this dictionary).
+    """
+
+    CHARACTERISTICS = (
+        coordination.Characteristics.NON_TIMEOUT_BASED,
+        coordination.Characteristics.DISTRIBUTED_ACROSS_THREADS,
+        coordination.Characteristics.DISTRIBUTED_ACROSS_PROCESSES,
+        coordination.Characteristics.DISTRIBUTED_ACROSS_HOSTS,
+        # Writes *always* go through a single leader process, but it may
+        # take a while for those writes to propagate to followers (and =
+        # during this time clients can read older values)...
+        coordination.Characteristics.SEQUENTIAL,
+    )
+    """
+    Tuple of :py:class:`~tooz.coordination.Characteristics` introspectable
+    enum member(s) that can be used to interogate how this driver works.
+    """
+
     def __init__(self, member_id, parsed_url, options):
-        super(BaseZooKeeperDriver, self).__init__()
+        super(KazooDriver, self).__init__(member_id, parsed_url, options)
         options = utils.collapse(options, exclude=['hosts'])
-        self._options = options
-        self._member_id = member_id
         self.timeout = int(options.get('timeout', '10'))
         self._namespace = options.get('namespace', self.TOOZ_NAMESPACE)
+        self._coord = self._make_client(parsed_url, options)
+        self._timeout_exception = self._coord.handler.timeout_exception
 
     def _start(self):
         try:
             self._coord.start(timeout=self.timeout)
         except self._coord.handler.timeout_exception as e:
             e_msg = encodeutils.exception_to_unicode(e)
-            coordination.raise_with_cause(coordination.ToozConnectionError,
-                                          "Operational error: %s" % e_msg,
-                                          cause=e)
+            utils.raise_with_cause(coordination.ToozConnectionError,
+                                   "Operational error: %s" % e_msg,
+                                   cause=e)
         try:
             self._coord.ensure_path(self._paths_join("/", self._namespace))
         except exceptions.KazooException as e:
             e_msg = encodeutils.exception_to_unicode(e)
-            coordination.raise_with_cause(coordination.ToozError,
-                                          "Operational error: %s" % e_msg,
-                                          cause=e)
-        self._group_members = collections.defaultdict(set)
-        self._watchers = collections.deque()
+            utils.raise_with_cause(tooz.ToozError,
+                                   "Operational error: %s" % e_msg,
+                                   cause=e)
         self._leader_locks = {}
 
     def _stop(self):
@@ -124,20 +178,20 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
-            coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(coordination.OperationTimedOut,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
         except exceptions.NodeExistsError:
             raise coordination.GroupAlreadyExist(group_id)
         except exceptions.NoNodeError as e:
-            coordination.raise_with_cause(coordination.ToozError,
-                                          "Tooz namespace '%s' has not"
-                                          " been created" % self._namespace,
-                                          cause=e)
+            utils.raise_with_cause(tooz.ToozError,
+                                   "Tooz namespace '%s' has not"
+                                   " been created" % self._namespace,
+                                   cause=e)
         except exceptions.ZookeeperError as e:
-            coordination.raise_with_cause(coordination.ToozError,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(tooz.ToozError,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
 
     def create_group(self, group_id):
         group_path = self._path_group(group_id)
@@ -152,17 +206,17 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
-            coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(coordination.OperationTimedOut,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
         except exceptions.NoNodeError:
             raise coordination.GroupNotCreated(group_id)
         except exceptions.NotEmptyError:
             raise coordination.GroupNotEmpty(group_id)
         except exceptions.ZookeeperError as e:
-            coordination.raise_with_cause(coordination.ToozError,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(tooz.ToozError,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
 
     def delete_group(self, group_id):
         group_path = self._path_group(group_id)
@@ -177,17 +231,17 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
-            coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(coordination.OperationTimedOut,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
         except exceptions.NodeExistsError:
             raise coordination.MemberAlreadyExist(group_id, member_id)
         except exceptions.NoNodeError:
             raise coordination.GroupNotCreated(group_id)
         except exceptions.ZookeeperError as e:
-            coordination.raise_with_cause(coordination.ToozError,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(tooz.ToozError,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
 
     def join_group(self, group_id, capabilities=b""):
         member_path = self._path_member(group_id, self._member_id)
@@ -205,15 +259,15 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
-            coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(coordination.OperationTimedOut,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
         except exceptions.NoNodeError:
             raise coordination.MemberNotJoined(group_id, member_id)
         except exceptions.ZookeeperError as e:
-            coordination.raise_with_cause(coordination.ToozError,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(tooz.ToozError,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
 
     def heartbeat(self):
         # Just fetch the base path (and do nothing with it); this will
@@ -223,15 +277,16 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             self._coord.get(base_path)
         except self._timeout_exception as e:
-            coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(coordination.OperationTimedOut,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
         except exceptions.NoNodeError:
             pass
         except exceptions.ZookeeperError as e:
-            coordination.raise_with_cause(coordination.ToozError,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(tooz.ToozError,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
+        return self.timeout
 
     def leave_group(self, group_id):
         member_path = self._path_member(group_id, self._member_id)
@@ -246,15 +301,15 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             members_ids = async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
-            coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(coordination.OperationTimedOut,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
         except exceptions.NoNodeError:
             raise coordination.GroupNotCreated(group_id)
         except exceptions.ZookeeperError as e:
-            coordination.raise_with_cause(coordination.ToozError,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(tooz.ToozError,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
         else:
             return set(m.encode('ascii') for m in members_ids)
 
@@ -271,15 +326,15 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
-            coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(coordination.OperationTimedOut,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
         except exceptions.NoNodeError:
             raise coordination.MemberNotJoined(group_id, member_id)
         except exceptions.ZookeeperError as e:
-            coordination.raise_with_cause(coordination.ToozError,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(tooz.ToozError,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
 
     def update_capabilities(self, group_id, capabilities):
         member_path = self._path_member(group_id, self._member_id)
@@ -296,15 +351,15 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             capabilities = async_result.get(block=True, timeout=timeout)[0]
         except timeout_exception as e:
-            coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(coordination.OperationTimedOut,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
         except exceptions.NoNodeError:
             raise coordination.MemberNotJoined(group_id, member_id)
         except exceptions.ZookeeperError as e:
-            coordination.raise_with_cause(coordination.ToozError,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(tooz.ToozError,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
         else:
             return cls._loads(capabilities)
 
@@ -324,15 +379,15 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
             capabilities, znode_stats = async_result.get(block=True,
                                                          timeout=timeout)
         except timeout_exception as e:
-            coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(coordination.OperationTimedOut,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
         except exceptions.NoNodeError:
             raise coordination.MemberNotJoined(group_id, member_id)
         except exceptions.ZookeeperError as e:
-            coordination.raise_with_cause(coordination.ToozError,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(tooz.ToozError,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
         else:
             member_info = {
                 'capabilities': cls._loads(capabilities),
@@ -353,18 +408,18 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         try:
             group_ids = async_result.get(block=True, timeout=timeout)
         except timeout_exception as e:
-            coordination.raise_with_cause(coordination.OperationTimedOut,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(coordination.OperationTimedOut,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
         except exceptions.NoNodeError as e:
-            coordination.raise_with_cause(coordination.ToozError,
-                                          "Tooz namespace '%s' has not"
-                                          " been created" % self._namespace,
-                                          cause=e)
+            utils.raise_with_cause(tooz.ToozError,
+                                   "Tooz namespace '%s' has not"
+                                   " been created" % self._namespace,
+                                   cause=e)
         except exceptions.ZookeeperError as e:
-            coordination.raise_with_cause(coordination.ToozError,
-                                          encodeutils.exception_to_unicode(e),
-                                          cause=e)
+            utils.raise_with_cause(tooz.ToozError,
+                                   encodeutils.exception_to_unicode(e),
+                                   cause=e)
         else:
             return set(g.encode('ascii') for g in group_ids)
 
@@ -387,78 +442,32 @@ class BaseZooKeeperDriver(coordination.CoordinationDriver):
         args.extend(more_args)
         cleaned_args = []
         for arg in args:
-            if isinstance(arg, six.binary_type):
-                cleaned_args.append(arg.decode('ascii'))
+            if isinstance(arg, bytes):
+                cleaned_args.append(
+                  encodeutils.safe_decode(arg, incoming='ascii')
+                )
             else:
                 cleaned_args.append(arg)
         return paths.join(*cleaned_args)
 
-
-class KazooDriver(BaseZooKeeperDriver):
-    """This driver uses the `kazoo`_ client against real `zookeeper`_ servers.
-
-    It **is** fully functional and implements all of the coordination
-    driver API(s). It stores data into `zookeeper`_ using znodes
-    and `msgpack`_ encoded values.
-
-    To configure the client to your liking a subset of the options defined at
-    http://kazoo.readthedocs.org/en/latest/api/client.html
-    will be extracted from the coordinator url (or any provided options),
-    so that a specific coordinator can be created that will work for you.
-
-    Currently the following options will be proxied to the contained client:
-
-    ================  ===============================  ====================
-    Name              Source                           Default
-    ================  ===============================  ====================
-    hosts             url netloc + 'hosts' option key  localhost:2181
-    timeout           'timeout' options key            10.0 (kazoo default)
-    connection_retry  'connection_retry' options key   None
-    command_retry     'command_retry' options key      None
-    randomize_hosts   'randomize_hosts' options key    True
-    ================  ===============================  ====================
-
-    .. _kazoo: http://kazoo.readthedocs.org/
-    .. _zookeeper: http://zookeeper.apache.org/
-    .. _msgpack: http://msgpack.org/
-    """
-
-    HANDLERS = {
-        'eventlet': eventlet_handler.SequentialEventletHandler,
-        'threading': threading_handler.SequentialThreadingHandler,
-    }
-    """
-    Restricted immutable dict of handler 'kinds' -> handler classes that
-    this driver can accept via 'handler' option key (the expected value for
-    this option is one of the keys in this dictionary).
-    """
-
-    CHARACTERISTICS = (
-        coordination.Characteristics.DISTRIBUTED_ACROSS_THREADS,
-        coordination.Characteristics.DISTRIBUTED_ACROSS_PROCESSES,
-        coordination.Characteristics.DISTRIBUTED_ACROSS_HOSTS,
-        # Writes *always* go through a single leader process, but it may
-        # take a while for those writes to propagate to followers (and =
-        # during this time clients can read older values)...
-        coordination.Characteristics.SEQUENTIAL,
-    )
-    """
-    Tuple of :py:class:`~tooz.coordination.Characteristics` introspectable
-    enum member(s) that can be used to interogate how this driver works.
-    """
-
-    def __init__(self, member_id, parsed_url, options):
-        super(KazooDriver, self).__init__(member_id, parsed_url, options)
-        self._coord = self._make_client(parsed_url, self._options)
-        self._member_id = member_id
-        self._timeout_exception = self._coord.handler.timeout_exception
-
     def _make_client(self, parsed_url, options):
         # Creates a kazoo client,
         # See: https://github.com/python-zk/kazoo/blob/2.2.1/kazoo/client.py
         # for what options a client takes...
+        if parsed_url.username and parsed_url.password:
+            username = parsed_url.username
+            password = parsed_url.password
+
+            digest_auth = "%s:%s" % (username, password)
+            digest_acl = security.make_digest_acl(username, password, all=True)
+            default_acl = (digest_acl,)
+            auth_data = [('digest', digest_auth)]
+        else:
+            default_acl = None
+            auth_data = None
+
         maybe_hosts = [parsed_url.netloc] + list(options.get('hosts', []))
-        hosts = list(compat_filter(None, maybe_hosts))
+        hosts = list(filter(None, maybe_hosts))
         if not hosts:
             hosts = ['localhost:2181']
         randomize_hosts = options.get('randomize_hosts', True)
@@ -468,6 +477,8 @@ class KazooDriver(BaseZooKeeperDriver):
             'connection_retry': options.get('connection_retry'),
             'command_retry': options.get('command_retry'),
             'randomize_hosts': strutils.bool_from_string(randomize_hosts),
+            'auth_data': auth_data,
+            'default_acl': default_acl,
         }
         handler_kind = options.get('handler')
         if handler_kind:
@@ -481,99 +492,6 @@ class KazooDriver(BaseZooKeeperDriver):
             client_kwargs['handler'] = handler_cls()
         return client.KazooClient(**client_kwargs)
 
-    def _watch_group(self, group_id):
-        get_members_req = self.get_members(group_id)
-
-        def on_children_change(children):
-            # If we don't have any hook, stop watching
-            if not self._has_hooks_for_group(group_id):
-                return False
-            children = set(children)
-            last_children = self._group_members[group_id]
-
-            for member_id in (children - last_children):
-                # Copy function in case it's removed later from the
-                # hook list
-                hooks = copy.copy(self._hooks_join_group[group_id])
-                self._watchers.append(
-                    lambda: hooks.run(
-                        coordination.MemberJoinedGroup(
-                            group_id,
-                            utils.to_binary(member_id))))
-
-            for member_id in (last_children - children):
-                # Copy function in case it's removed later from the
-                # hook list
-                hooks = copy.copy(self._hooks_leave_group[group_id])
-                self._watchers.append(
-                    lambda: hooks.run(
-                        coordination.MemberLeftGroup(
-                            group_id,
-                            utils.to_binary(member_id))))
-
-            self._group_members[group_id] = children
-
-        # Initialize the current member list
-        self._group_members[group_id] = get_members_req.get()
-
-        try:
-            self._coord.ChildrenWatch(self._path_group(group_id),
-                                      on_children_change)
-        except exceptions.NoNodeError:
-            raise coordination.GroupNotCreated(group_id)
-
-    def watch_join_group(self, group_id, callback):
-        # Check if we already have hooks for this group_id, if not, start
-        # watching it.
-        already_being_watched = self._has_hooks_for_group(group_id)
-
-        # Add the hook before starting watching to avoid race conditions
-        # as the watching executor can be in a thread
-        super(BaseZooKeeperDriver, self).watch_join_group(
-            group_id, callback)
-
-        if not already_being_watched:
-            try:
-                self._watch_group(group_id)
-            except Exception:
-                # Rollback and unregister the hook
-                self.unwatch_join_group(group_id, callback)
-                raise
-
-    def unwatch_join_group(self, group_id, callback):
-        return super(BaseZooKeeperDriver, self).unwatch_join_group(
-            group_id, callback)
-
-    def watch_leave_group(self, group_id, callback):
-        # Check if we already have hooks for this group_id, if not, start
-        # watching it.
-        already_being_watched = self._has_hooks_for_group(group_id)
-
-        # Add the hook before starting watching to avoid race conditions
-        # as the watching executor can be in a thread
-        super(BaseZooKeeperDriver, self).watch_leave_group(
-            group_id, callback)
-
-        if not already_being_watched:
-            try:
-                self._watch_group(group_id)
-            except Exception:
-                # Rollback and unregister the hook
-                self.unwatch_leave_group(group_id, callback)
-                raise
-
-    def unwatch_leave_group(self, group_id, callback):
-        return super(BaseZooKeeperDriver, self).unwatch_leave_group(
-            group_id, callback)
-
-    def watch_elected_as_leader(self, group_id, callback):
-        return super(BaseZooKeeperDriver, self).watch_elected_as_leader(
-            group_id, callback)
-
-    def unwatch_elected_as_leader(self, group_id, callback):
-        return super(BaseZooKeeperDriver, self).unwatch_elected_as_leader(
-            group_id, callback)
-
     def stand_down_group_leader(self, group_id):
         if group_id in self._leader_locks:
             self._leader_locks[group_id].release()
@@ -584,7 +502,7 @@ class KazooDriver(BaseZooKeeperDriver):
         if group_id not in self._leader_locks:
             self._leader_locks[group_id] = self._coord.Lock(
                 self._path_group(group_id) + "/leader",
-                self._member_id.decode('ascii'))
+                encodeutils.safe_decode(self._member_id, incoming='ascii'))
         return self._leader_locks[group_id]
 
     def get_leader(self, group_id):
@@ -598,11 +516,11 @@ class KazooDriver(BaseZooKeeperDriver):
     def get_lock(self, name):
         z_lock = self._coord.Lock(
             self._paths_join(b"/", self._namespace, b"locks", name),
-            self._member_id.decode('ascii'))
+            encodeutils.safe_decode(self._member_id, incoming='ascii'))
         return ZooKeeperLock(name, z_lock)
 
     def run_elect_coordinator(self):
-        for group_id in six.iterkeys(self._hooks_elected_leader):
+        for group_id in self._hooks_elected_leader.keys():
             leader_lock = self._get_group_leader_lock(group_id)
             if leader_lock.is_acquired:
                 # Previously acquired/still leader, leave it be...
@@ -615,10 +533,7 @@ class KazooDriver(BaseZooKeeperDriver):
                         self._member_id))
 
     def run_watchers(self, timeout=None):
-        results = []
-        while self._watchers:
-            cb = self._watchers.popleft()
-            results.extend(cb())
+        results = super(KazooDriver, self).run_watchers(timeout)
         self.run_elect_coordinator()
         return results
 
@@ -629,7 +544,7 @@ class ZooAsyncResult(coordination.CoordAsyncResult):
         self._handler = handler
         self._kwargs = kwargs
 
-    def get(self, timeout=10):
+    def get(self, timeout=None):
         return self._handler(self._kazoo_async_result, timeout, **self._kwargs)
 
     def done(self):
