@@ -15,13 +15,19 @@
 #    under the License.
 
 import functools
+import os
 
-import six
+import fixtures
+from oslo_utils import uuidutils
 from testtools import testcase
 
 import tooz
 
 
+def get_random_uuid():
+    return uuidutils.generate_uuid().encode('ascii')
+
+
 def _skip_decorator(func):
     @functools.wraps(func)
     def skip_if_not_implemented(*args, **kwargs):
@@ -42,6 +48,24 @@ class SkipNotImplementedMeta(type):
         return type.__new__(cls, name, bases, local)
 
 
-@six.add_metaclass(SkipNotImplementedMeta)
-class TestCaseSkipNotImplemented(testcase.TestCase):
-    pass
+class TestWithCoordinator(testcase.TestCase, metaclass=SkipNotImplementedMeta):
+    url = os.getenv("TOOZ_TEST_URL")
+
+    def setUp(self):
+        super(TestWithCoordinator, self).setUp()
+        if self.url is None:
+            raise RuntimeError("No URL set for this driver")
+        if os.getenv("TOOZ_TEST_ETCD3"):
+            self.url = self.url.replace("etcd://", "etcd3://")
+        if os.getenv("TOOZ_TEST_ETCD3GW"):
+            self.url = self.url.replace("etcd://", "etcd3+http://")
+        self.useFixture(fixtures.NestedTempfile())
+        self.group_id = get_random_uuid()
+        self.member_id = get_random_uuid()
+        self._coord = tooz.coordination.get_coordinator(self.url,
+                                                        self.member_id)
+        self._coord.start()
+
+    def tearDown(self):
+        self._coord.stop()
+        super(TestWithCoordinator, self).tearDown()
