@@ -14,18 +14,16 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-import os
 import threading
 import time
-import uuid
+from unittest import mock
+import urllib
 
 from concurrent import futures
-import fixtures
-import mock
-from six.moves.urllib import parse
 from testtools import matchers
 from testtools import testcase
 
+import tooz
 import tooz.coordination
 from tooz import tests
 
@@ -39,10 +37,7 @@ def try_to_lock_job(name, coord, url, member_id):
     return lock2.acquire(blocking=False)
 
 
-class TestAPI(tests.TestCaseSkipNotImplemented):
-
-    url = os.getenv("TOOZ_TEST_URL")
-
+class TestAPI(tests.TestWithCoordinator):
     def assertRaisesAny(self, exc_classes, callable_obj, *args, **kwargs):
         checkers = [matchers.MatchesException(exc_class)
                     for exc_class in exc_classes]
@@ -50,26 +45,11 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         callable_obj = testcase.Nullary(callable_obj, *args, **kwargs)
         self.assertThat(callable_obj, matcher)
 
-    def setUp(self):
-        super(TestAPI, self).setUp()
-        if self.url is None:
-            self.skipTest("No URL set for this driver")
-        self.useFixture(fixtures.NestedTempfile())
-        self.group_id = self._get_random_uuid()
-        self.member_id = self._get_random_uuid()
-        self._coord = tooz.coordination.get_coordinator(self.url,
-                                                        self.member_id)
-        self._coord.start()
-
-    def tearDown(self):
-        self._coord.stop()
-        super(TestAPI, self).tearDown()
-
     def test_connection_error_bad_host(self):
         if (tooz.coordination.Characteristics.DISTRIBUTED_ACROSS_HOSTS
            not in self._coord.CHARACTERISTICS):
             self.skipTest("This driver is not distributed across hosts")
-        scheme = parse.urlparse(self.url).scheme
+        scheme = urllib.parse.urlparse(self.url).scheme
         coord = tooz.coordination.get_coordinator(
             "%s://localhost:1/f00" % scheme,
             self.member_id)
@@ -79,17 +59,17 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
     def test_stop_first(self):
         c = tooz.coordination.get_coordinator(self.url,
                                               self.member_id)
-        self.assertRaises(tooz.coordination.ToozError,
+        self.assertRaises(tooz.ToozError,
                           c.stop)
 
     def test_create_group(self):
         self._coord.create_group(self.group_id).get()
         all_group_ids = self._coord.get_groups().get()
-        self.assertTrue(self.group_id in all_group_ids)
+        self.assertIn(self.group_id, all_group_ids)
 
     def test_get_lock_release_broken(self):
-        name = self._get_random_uuid()
-        memberid2 = self._get_random_uuid()
+        name = tests.get_random_uuid()
+        memberid2 = tests.get_random_uuid()
         coord2 = tooz.coordination.get_coordinator(self.url,
                                                    memberid2)
         coord2.start()
@@ -101,7 +81,7 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         self.assertTrue(lock2.acquire(blocking=False))
         self.assertFalse(lock1.release())
         # Assert lock is not accidentally broken now
-        memberid3 = self._get_random_uuid()
+        memberid3 = tests.get_random_uuid()
         coord3 = tooz.coordination.get_coordinator(self.url,
                                                    memberid3)
         coord3.start()
@@ -115,7 +95,7 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
                           create_group.get)
 
     def test_get_groups(self):
-        groups_ids = [self._get_random_uuid() for _ in range(0, 5)]
+        groups_ids = [tests.get_random_uuid() for _ in range(0, 5)]
         for group_id in groups_ids:
             self._coord.create_group(group_id).get()
         created_groups = self._coord.get_groups().get()
@@ -125,10 +105,10 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
     def test_delete_group(self):
         self._coord.create_group(self.group_id).get()
         all_group_ids = self._coord.get_groups().get()
-        self.assertTrue(self.group_id in all_group_ids)
+        self.assertIn(self.group_id, all_group_ids)
         self._coord.delete_group(self.group_id).get()
         all_group_ids = self._coord.get_groups().get()
-        self.assertFalse(self.group_id in all_group_ids)
+        self.assertNotIn(self.group_id, all_group_ids)
         join_group = self._coord.join_group(self.group_id)
         self.assertRaises(tooz.coordination.GroupNotCreated,
                           join_group.get)
@@ -144,20 +124,25 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         delete = self._coord.delete_group(self.group_id)
         self.assertRaises(tooz.coordination.GroupNotEmpty,
                           delete.get)
-        self._coord.leave_group(self.group_id)
+        self._coord.leave_group(self.group_id).get()
         self._coord.delete_group(self.group_id).get()
 
     def test_join_group(self):
         self._coord.create_group(self.group_id).get()
         self._coord.join_group(self.group_id).get()
         member_list = self._coord.get_members(self.group_id).get()
-        self.assertTrue(self.member_id in member_list)
+        self.assertIn(self.member_id, member_list)
 
     def test_join_nonexistent_group(self):
         join_group = self._coord.join_group(self.group_id)
         self.assertRaises(tooz.coordination.GroupNotCreated,
                           join_group.get)
 
+    def test_join_group_create(self):
+        self._coord.join_group_create(self.group_id)
+        member_list = self._coord.get_members(self.group_id).get()
+        self.assertIn(self.member_id, member_list)
+
     def test_join_group_with_member_id_already_exists(self):
         self._coord.create_group(self.group_id).get()
         self._coord.join_group(self.group_id).get()
@@ -171,20 +156,20 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
     def test_leave_group(self):
         self._coord.create_group(self.group_id).get()
         all_group_ids = self._coord.get_groups().get()
-        self.assertTrue(self.group_id in all_group_ids)
+        self.assertIn(self.group_id, all_group_ids)
         self._coord.join_group(self.group_id).get()
         member_list = self._coord.get_members(self.group_id).get()
-        self.assertTrue(self.member_id in member_list)
+        self.assertIn(self.member_id, member_list)
         member_ids = self._coord.get_members(self.group_id).get()
-        self.assertTrue(self.member_id in member_ids)
+        self.assertIn(self.member_id, member_ids)
         self._coord.leave_group(self.group_id).get()
         new_member_objects = self._coord.get_members(self.group_id).get()
         new_member_list = [member.member_id for member in new_member_objects]
-        self.assertTrue(self.member_id not in new_member_list)
+        self.assertNotIn(self.member_id, new_member_list)
 
     def test_leave_nonexistent_group(self):
         all_group_ids = self._coord.get_groups().get()
-        self.assertTrue(self.group_id not in all_group_ids)
+        self.assertNotIn(self.group_id, all_group_ids)
         leave_group = self._coord.leave_group(self.group_id)
         # Drivers raise one of those depending on their capability
         self.assertRaisesAny([tooz.coordination.MemberNotJoined,
@@ -194,13 +179,13 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
     def test_leave_group_not_joined_by_member(self):
         self._coord.create_group(self.group_id).get()
         all_group_ids = self._coord.get_groups().get()
-        self.assertTrue(self.group_id in all_group_ids)
+        self.assertIn(self.group_id, all_group_ids)
         leave_group = self._coord.leave_group(self.group_id)
         self.assertRaises(tooz.coordination.MemberNotJoined,
                           leave_group.get)
 
     def test_get_lock_twice_locked_one_released_two(self):
-        name = self._get_random_uuid()
+        name = tests.get_random_uuid()
         lock1 = self._coord.get_lock(name)
         lock2 = self._coord.get_lock(name)
         self.assertTrue(lock1.acquire())
@@ -210,8 +195,8 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         self.assertFalse(lock2.release())
 
     def test_get_members(self):
-        group_id_test2 = self._get_random_uuid()
-        member_id_test2 = self._get_random_uuid()
+        group_id_test2 = tests.get_random_uuid()
+        member_id_test2 = tests.get_random_uuid()
         client2 = tooz.coordination.get_coordinator(self.url,
                                                     member_id_test2)
         client2.start()
@@ -220,8 +205,7 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         self._coord.join_group(group_id_test2).get()
         client2.join_group(group_id_test2).get()
         members_ids = self._coord.get_members(group_id_test2).get()
-        self.assertTrue(self.member_id in members_ids)
-        self.assertTrue(member_id_test2 in members_ids)
+        self.assertEqual({self.member_id, member_id_test2}, members_ids)
 
     def test_get_member_capabilities(self):
         self._coord.create_group(self.group_id).get()
@@ -237,7 +221,7 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
             'type': 'warrior',
             'abilities': ['fight', 'flight', 'double-hit-damage'],
         }
-        self._coord.join_group(self.group_id, caps)
+        self._coord.join_group(self.group_id, caps).get()
         capa = self._coord.get_member_capabilities(self.group_id,
                                                    self.member_id).get()
         self.assertEqual(capa, caps)
@@ -260,7 +244,7 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
 
     def test_get_member_info(self):
         self._coord.create_group(self.group_id).get()
-        self._coord.join_group(self.group_id, b"test_capabilities")
+        self._coord.join_group(self.group_id, b"test_capabilities").get()
 
         member_info = self._coord.get_member_info(self.group_id,
                                                   self.member_id).get()
@@ -275,7 +259,7 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         member_info = {'capabilities': 'caps',
                        'created_at': '0',
                        'updated_at': '0'}
-        self._coord.join_group(self.group_id, caps)
+        self._coord.join_group(self.group_id, caps).get()
         member_info = self._coord.get_member_info(self.group_id,
                                                   self.member_id).get()
         self.assertEqual(member_info['capabilities'], caps)
@@ -290,7 +274,7 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
 
     def test_get_member_info_nonjoined_member(self):
         self._coord.create_group(self.group_id).get()
-        member_id = self._get_random_uuid()
+        member_id = tests.get_random_uuid()
         member_info = self._coord.get_member_info(self.group_id,
                                                   member_id)
         self.assertRaises(tooz.coordination.MemberNotJoined,
@@ -342,7 +326,7 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
             heart.wait()
 
     def test_disconnect_leave_group(self):
-        member_id_test2 = self._get_random_uuid()
+        member_id_test2 = tests.get_random_uuid()
         client2 = tooz.coordination.get_coordinator(self.url,
                                                     member_id_test2)
         client2.start()
@@ -350,12 +334,12 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         self._coord.join_group(self.group_id).get()
         client2.join_group(self.group_id).get()
         members_ids = self._coord.get_members(self.group_id).get()
-        self.assertTrue(self.member_id in members_ids)
-        self.assertTrue(member_id_test2 in members_ids)
+        self.assertIn(self.member_id, members_ids)
+        self.assertIn(member_id_test2, members_ids)
         client2.stop()
         members_ids = self._coord.get_members(self.group_id).get()
-        self.assertTrue(self.member_id in members_ids)
-        self.assertTrue(member_id_test2 not in members_ids)
+        self.assertIn(self.member_id, members_ids)
+        self.assertNotIn(member_id_test2, members_ids)
 
     def test_timeout(self):
         if (tooz.coordination.Characteristics.NON_TIMEOUT_BASED
@@ -370,20 +354,24 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         self._coord = tooz.coordination.get_coordinator(url, self.member_id)
         self._coord.start()
 
-        member_id_test2 = self._get_random_uuid()
+        member_id_test2 = tests.get_random_uuid()
         client2 = tooz.coordination.get_coordinator(url, member_id_test2)
         client2.start()
         self._coord.create_group(self.group_id).get()
         self._coord.join_group(self.group_id).get()
         client2.join_group(self.group_id).get()
         members_ids = self._coord.get_members(self.group_id).get()
-        self.assertTrue(self.member_id in members_ids)
-        self.assertTrue(member_id_test2 in members_ids)
+        self.assertIn(self.member_id, members_ids)
+        self.assertIn(member_id_test2, members_ids)
 
         # Watch the group, we want to be sure that when client2 is kicked out
         # we get an event.
         self._coord.watch_leave_group(self.group_id, self._set_event)
 
+        # Run watchers to be sure we initialize the member cache and we *know*
+        # client2 is a member now
+        self._coord.run_watchers()
+
         time.sleep(3)
         self._coord.heartbeat()
         time.sleep(3)
@@ -394,22 +382,24 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         while True:
             if self._coord.run_watchers():
                 break
-        self.assertTrue(self.member_id in members_ids)
-        self.assertTrue(member_id_test2 not in members_ids)
+        self.assertIn(self.member_id, members_ids)
+        self.assertNotIn(member_id_test2, members_ids)
         # Check that the event has been triggered
-        self.assertIsInstance(self.event,
-                              tooz.coordination.MemberLeftGroup)
-        self.assertEqual(member_id_test2,
-                         self.event.member_id)
-        self.assertEqual(self.group_id,
-                         self.event.group_id)
+        self.assertEqual(1, len(self.events))
+        event = self.events[0]
+        self.assertIsInstance(event, tooz.coordination.MemberLeftGroup)
+        self.assertEqual(member_id_test2, event.member_id)
+        self.assertEqual(self.group_id, event.group_id)
 
     def _set_event(self, event):
-        self.event = event
+        if not hasattr(self, "events"):
+            self.events = [event]
+        else:
+            self.events.append(event)
         return 42
 
     def test_watch_group_join(self):
-        member_id_test2 = self._get_random_uuid()
+        member_id_test2 = tests.get_random_uuid()
         client2 = tooz.coordination.get_coordinator(self.url,
                                                     member_id_test2)
         client2.start()
@@ -421,29 +411,28 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         # Join the group
         client2.join_group(self.group_id).get()
         members_ids = self._coord.get_members(self.group_id).get()
-        self.assertTrue(member_id_test2 in members_ids)
+        self.assertIn(member_id_test2, members_ids)
         while True:
             if self._coord.run_watchers():
                 break
-        self.assertIsInstance(self.event,
-                              tooz.coordination.MemberJoinedGroup)
-        self.assertEqual(member_id_test2,
-                         self.event.member_id)
-        self.assertEqual(self.group_id,
-                         self.event.group_id)
+        self.assertEqual(1, len(self.events))
+        event = self.events[0]
+        self.assertIsInstance(event, tooz.coordination.MemberJoinedGroup)
+        self.assertEqual(member_id_test2, event.member_id)
+        self.assertEqual(self.group_id, event.group_id)
 
         # Stop watching
         self._coord.unwatch_join_group(self.group_id, self._set_event)
-        self.event = None
+        self.events = []
 
         # Leave and rejoin group
         client2.leave_group(self.group_id).get()
         client2.join_group(self.group_id).get()
         self._coord.run_watchers()
-        self.assertIsNone(self.event)
+        self.assertEqual([], self.events)
 
     def test_watch_leave_group(self):
-        member_id_test2 = self._get_random_uuid()
+        member_id_test2 = tests.get_random_uuid()
         client2 = tooz.coordination.get_coordinator(self.url,
                                                     member_id_test2)
         client2.start()
@@ -468,22 +457,21 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
             if self._coord.run_watchers():
                 break
 
-        self.assertIsInstance(self.event,
-                              tooz.coordination.MemberLeftGroup)
-        self.assertEqual(member_id_test2,
-                         self.event.member_id)
-        self.assertEqual(self.group_id,
-                         self.event.group_id)
+        self.assertEqual(1, len(self.events))
+        event = self.events[0]
+        self.assertIsInstance(event, tooz.coordination.MemberLeftGroup)
+        self.assertEqual(member_id_test2, event.member_id)
+        self.assertEqual(self.group_id, event.group_id)
 
         # Stop watching
         self._coord.unwatch_leave_group(self.group_id, self._set_event)
-        self.event = None
+        self.events = []
 
         # Rejoin and releave group
         client2.join_group(self.group_id).get()
         client2.leave_group(self.group_id).get()
         self._coord.run_watchers()
-        self.assertIsNone(self.event)
+        self.assertEqual([], self.events)
 
     def test_watch_join_group_disappear(self):
         if not hasattr(self._coord, '_destroy_group'):
@@ -494,7 +482,7 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         self._coord.watch_join_group(self.group_id, self._set_event)
         self._coord.watch_leave_group(self.group_id, self._set_event)
 
-        member_id_test2 = self._get_random_uuid()
+        member_id_test2 = tests.get_random_uuid()
         client2 = tooz.coordination.get_coordinator(self.url,
                                                     member_id_test2)
         client2.start()
@@ -503,9 +491,10 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         while True:
             if self._coord.run_watchers():
                 break
-        self.assertIsInstance(self.event,
-                              tooz.coordination.MemberJoinedGroup)
-        self.event = None
+        self.assertEqual(1, len(self.events))
+        event = self.events[0]
+        self.assertIsInstance(event, tooz.coordination.MemberJoinedGroup)
+        self.events = []
 
         # Force the group to disappear...
         self._coord._destroy_group(self.group_id)
@@ -514,8 +503,9 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
             if self._coord.run_watchers():
                 break
 
-        self.assertIsInstance(self.event,
-                              tooz.coordination.MemberLeftGroup)
+        self.assertEqual(1, len(self.events))
+        event = self.events[0]
+        self.assertIsInstance(event, tooz.coordination.MemberLeftGroup)
 
     def test_watch_join_group_non_existent(self):
         self.assertRaises(tooz.coordination.GroupNotCreated,
@@ -530,7 +520,7 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         self._coord.watch_join_group(self.group_id, self._set_event)
         self._coord.watch_leave_group(self.group_id, self._set_event)
 
-        member_id_test2 = self._get_random_uuid()
+        member_id_test2 = tests.get_random_uuid()
         client2 = tooz.coordination.get_coordinator(self.url,
                                                     member_id_test2)
         client2.start()
@@ -547,7 +537,9 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         # Only works for clients that have access to the groups they are part
         # of, to ensure that after we got booted out by client3 that this
         # client now no longer believes its part of the group.
-        if hasattr(self._coord, '_joined_groups'):
+        if (hasattr(self._coord, '_joined_groups')
+           and (self._coord.run_watchers
+                == tooz.coordination.CoordinationDriverCachedRunWatchers.run_watchers)):  # noqa
             self.assertIn(self.group_id, self._coord._joined_groups)
             self._coord.run_watchers()
             self.assertNotIn(self.group_id, self._coord._joined_groups)
@@ -564,45 +556,44 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         self._coord.watch_elected_as_leader(self.group_id, self._set_event)
         self._coord.run_watchers()
 
-        self.assertIsInstance(self.event,
-                              tooz.coordination.LeaderElected)
-        self.assertEqual(self.member_id,
-                         self.event.member_id)
-        self.assertEqual(self.group_id,
-                         self.event.group_id)
+        self.assertEqual(1, len(self.events))
+        event = self.events[0]
+        self.assertIsInstance(event, tooz.coordination.LeaderElected)
+        self.assertEqual(self.member_id, event.member_id)
+        self.assertEqual(self.group_id, event.group_id)
 
     def test_run_for_election_multiple_clients(self):
         self._coord.create_group(self.group_id).get()
         self._coord.watch_elected_as_leader(self.group_id, self._set_event)
         self._coord.run_watchers()
 
-        member_id_test2 = self._get_random_uuid()
+        member_id_test2 = tests.get_random_uuid()
         client2 = tooz.coordination.get_coordinator(self.url,
                                                     member_id_test2)
         client2.start()
         client2.watch_elected_as_leader(self.group_id, self._set_event)
         client2.run_watchers()
 
-        self.assertIsInstance(self.event,
-                              tooz.coordination.LeaderElected)
-        self.assertEqual(self.member_id,
-                         self.event.member_id)
-        self.assertEqual(self.group_id,
-                         self.event.group_id)
+        self.assertEqual(1, len(self.events))
+        event = self.events[0]
+        self.assertIsInstance(event, tooz.coordination.LeaderElected)
+        self.assertEqual(self.member_id, event.member_id)
+        self.assertEqual(self.group_id, event.group_id)
         self.assertEqual(self._coord.get_leader(self.group_id).get(),
                          self.member_id)
 
-        self.event = None
+        self.events = []
 
         self._coord.stop()
         client2.run_watchers()
 
-        self.assertIsInstance(self.event,
-                              tooz.coordination.LeaderElected)
+        self.assertEqual(1, len(self.events))
+        event = self.events[0]
+
+        self.assertIsInstance(event, tooz.coordination.LeaderElected)
         self.assertEqual(member_id_test2,
-                         self.event.member_id)
-        self.assertEqual(self.group_id,
-                         self.event.group_id)
+                         event.member_id)
+        self.assertEqual(self.group_id, event.group_id)
         self.assertEqual(client2.get_leader(self.group_id).get(),
                          member_id_test2)
 
@@ -613,12 +604,12 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         self._coord.create_group(self.group_id).get()
 
         leader = self._coord.get_leader(self.group_id).get()
-        self.assertEqual(leader, None)
+        self.assertIsNone(leader)
 
         self._coord.join_group(self.group_id).get()
 
         leader = self._coord.get_leader(self.group_id).get()
-        self.assertEqual(leader, None)
+        self.assertIsNone(leader)
 
         # Let's get elected
         self._coord.watch_elected_as_leader(self.group_id, self._set_event)
@@ -632,43 +623,43 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         self._coord.watch_elected_as_leader(self.group_id, self._set_event)
         self._coord.run_watchers()
 
-        member_id_test2 = self._get_random_uuid()
+        member_id_test2 = tests.get_random_uuid()
         client2 = tooz.coordination.get_coordinator(self.url,
                                                     member_id_test2)
         client2.start()
         client2.watch_elected_as_leader(self.group_id, self._set_event)
         client2.run_watchers()
 
-        self.assertIsInstance(self.event,
-                              tooz.coordination.LeaderElected)
-        self.assertEqual(self.member_id,
-                         self.event.member_id)
-        self.assertEqual(self.group_id,
-                         self.event.group_id)
+        self.assertEqual(1, len(self.events))
+        event = self.events[0]
 
-        self.event = None
+        self.assertIsInstance(event, tooz.coordination.LeaderElected)
+        self.assertEqual(self.member_id, event.member_id)
+        self.assertEqual(self.group_id, event.group_id)
+
+        self.events = []
 
         self._coord.stand_down_group_leader(self.group_id)
         client2.run_watchers()
 
-        self.assertIsInstance(self.event,
-                              tooz.coordination.LeaderElected)
-        self.assertEqual(member_id_test2,
-                         self.event.member_id)
-        self.assertEqual(self.group_id,
-                         self.event.group_id)
+        self.assertEqual(1, len(self.events))
+        event = self.events[0]
 
-        self.event = None
+        self.assertIsInstance(event, tooz.coordination.LeaderElected)
+        self.assertEqual(member_id_test2, event.member_id)
+        self.assertEqual(self.group_id, event.group_id)
+
+        self.events = []
 
         client2.stand_down_group_leader(self.group_id)
         self._coord.run_watchers()
 
-        self.assertIsInstance(self.event,
-                              tooz.coordination.LeaderElected)
-        self.assertEqual(self.member_id,
-                         self.event.member_id)
-        self.assertEqual(self.group_id,
-                         self.event.group_id)
+        self.assertEqual(1, len(self.events))
+        event = self.events[0]
+
+        self.assertIsInstance(event, tooz.coordination.LeaderElected)
+        self.assertEqual(self.member_id, event.member_id)
+        self.assertEqual(self.group_id, event.group_id)
 
     def test_unwatch_elected_as_leader(self):
         # Create a group and add a elected_as_leader callback
@@ -676,7 +667,8 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         self._coord.watch_elected_as_leader(self.group_id, self._set_event)
 
         # Ensure exactly one leader election hook exists
-        self.assertEqual(1, len(self._coord._hooks_elected_leader))
+        self.assertEqual(1,
+                         len(self._coord._hooks_elected_leader[self.group_id]))
 
         # Unwatch, and ensure no leader election hooks exist
         self._coord.unwatch_elected_as_leader(self.group_id, self._set_event)
@@ -694,6 +686,19 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
                           self._coord.unwatch_join_group,
                           self.group_id, lambda x: None)
 
+    def test_unwatch_leave_group(self):
+        # Create a group and add a leave_group callback
+        self._coord.create_group(self.group_id).get()
+        self.assertEqual(0, len(self._coord._hooks_leave_group))
+        self._coord.watch_leave_group(self.group_id, self._set_event)
+
+        # Ensure exactly one leave group hook exists
+        self.assertEqual(1, len(self._coord._hooks_leave_group[self.group_id]))
+
+        # Unwatch, and ensure no leave group hooks exist
+        self._coord.unwatch_leave_group(self.group_id, self._set_event)
+        self.assertEqual(0, len(self._coord._hooks_leave_group))
+
     def test_unwatch_leave_group_callback_not_found(self):
         self._coord.create_group(self.group_id).get()
         self.assertRaises(tooz.coordination.WatchCallbackNotFound,
@@ -701,14 +706,76 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
                           self.group_id, lambda x: None)
 
     def test_get_lock(self):
-        lock = self._coord.get_lock(self._get_random_uuid())
+        lock = self._coord.get_lock(tests.get_random_uuid())
         self.assertTrue(lock.acquire())
         self.assertTrue(lock.release())
         with lock:
             pass
 
+    def test_heartbeat_lock_not_acquired(self):
+        lock = self._coord.get_lock(tests.get_random_uuid())
+        # Not all locks need heartbeat
+        if hasattr(lock, "heartbeat"):
+            self.assertFalse(lock.heartbeat())
+
+    def test_get_shared_lock(self):
+        lock = self._coord.get_lock(tests.get_random_uuid())
+        self.assertTrue(lock.acquire(shared=True))
+        self.assertTrue(lock.release())
+        with lock(shared=True):
+            pass
+
+    def test_get_shared_lock_locking_same_lock_twice(self):
+        lock = self._coord.get_lock(tests.get_random_uuid())
+        self.assertTrue(lock.acquire(shared=True))
+        self.assertTrue(lock.acquire(shared=True))
+        self.assertTrue(lock.release())
+        self.assertTrue(lock.release())
+        self.assertFalse(lock.release())
+        with lock(shared=True):
+            pass
+
+    def test_get_shared_lock_locking_two_lock(self):
+        name = tests.get_random_uuid()
+        lock1 = self._coord.get_lock(name)
+        coord = tooz.coordination.get_coordinator(
+            self.url, tests.get_random_uuid())
+        coord.start()
+        lock2 = coord.get_lock(name)
+
+        self.assertTrue(lock1.acquire(shared=True))
+        self.assertTrue(lock2.acquire(shared=True))
+        self.assertTrue(lock1.release())
+        self.assertTrue(lock2.release())
+
+    def test_get_lock_locking_shared_and_exclusive(self):
+        name = tests.get_random_uuid()
+        lock1 = self._coord.get_lock(name)
+        coord = tooz.coordination.get_coordinator(
+            self.url, tests.get_random_uuid())
+        coord.start()
+        lock2 = coord.get_lock(name)
+
+        self.assertTrue(lock1.acquire(shared=True))
+        self.assertFalse(lock2.acquire(blocking=False))
+        self.assertTrue(lock1.release())
+        self.assertFalse(lock2.release())
+
+    def test_get_lock_locking_exclusive_and_shared(self):
+        name = tests.get_random_uuid()
+        lock1 = self._coord.get_lock(name)
+        coord = tooz.coordination.get_coordinator(
+            self.url, tests.get_random_uuid())
+        coord.start()
+        lock2 = coord.get_lock(name)
+
+        self.assertTrue(lock1.acquire())
+        self.assertFalse(lock2.acquire(shared=True, blocking=False))
+        self.assertTrue(lock1.release())
+        self.assertFalse(lock2.release())
+
     def test_get_lock_concurrency_locking_same_lock(self):
-        lock = self._coord.get_lock(self._get_random_uuid())
+        lock = self._coord.get_lock(tests.get_random_uuid())
 
         graceful_ending = threading.Event()
 
@@ -729,19 +796,47 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
 
     def _do_test_get_lock_concurrency_locking_two_lock(self, executor,
                                                        use_same_coord):
-        name = self._get_random_uuid()
+        name = tests.get_random_uuid()
         lock1 = self._coord.get_lock(name)
         with lock1:
             with executor(max_workers=1) as e:
                 coord = self._coord if use_same_coord else None
                 f = e.submit(try_to_lock_job, name, coord, self.url,
-                             self._get_random_uuid())
+                             tests.get_random_uuid())
                 self.assertFalse(f.result())
 
+    def _do_test_get_lock_serial_locking_two_lock(self, executor,
+                                                  use_same_coord):
+        name = tests.get_random_uuid()
+        lock1 = self._coord.get_lock(name)
+        lock1.acquire()
+        lock1.release()
+        with executor(max_workers=1) as e:
+            coord = self._coord if use_same_coord else None
+            f = e.submit(try_to_lock_job, name, coord, self.url,
+                         tests.get_random_uuid())
+            self.assertTrue(f.result())
+
     def test_get_lock_concurrency_locking_two_lock_process(self):
+        # NOTE(jd) Using gRPC and forking is not supported so this test might
+        # very likely hang forever or crash. See
+        # https://github.com/grpc/grpc/issues/10140#issuecomment-297548714 for
+        # more info.
+        if self.url.startswith("etcd3://"):
+            self.skipTest("Unable to use etcd3 with fork()")
         self._do_test_get_lock_concurrency_locking_two_lock(
             futures.ProcessPoolExecutor, False)
 
+    def test_get_lock_serial_locking_two_lock_process(self):
+        # NOTE(jd) Using gRPC and forking is not supported so this test might
+        # very likely hang forever or crash. See
+        # https://github.com/grpc/grpc/issues/10140#issuecomment-297548714 for
+        # more info.
+        if self.url.startswith("etcd3://"):
+            self.skipTest("Unable to use etcd3 with fork()")
+        self._do_test_get_lock_serial_locking_two_lock(
+            futures.ProcessPoolExecutor, False)
+
     def test_get_lock_concurrency_locking_two_lock_thread1(self):
         self._do_test_get_lock_concurrency_locking_two_lock(
             futures.ThreadPoolExecutor, False)
@@ -754,8 +849,8 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         # NOTE(sileht): some database based lock can have only
         # one lock per connection, this test ensures acquiring a
         # second lock doesn't release the first one.
-        lock1 = self._coord.get_lock(self._get_random_uuid())
-        lock2 = self._coord.get_lock(self._get_random_uuid())
+        lock1 = self._coord.get_lock(tests.get_random_uuid())
+        lock2 = self._coord.get_lock(tests.get_random_uuid())
 
         graceful_ending = threading.Event()
         thread_locked = threading.Event()
@@ -781,14 +876,14 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         self.assertTrue(graceful_ending.is_set())
 
     def test_get_lock_twice_locked_twice(self):
-        name = self._get_random_uuid()
+        name = tests.get_random_uuid()
         lock1 = self._coord.get_lock(name)
         lock2 = self._coord.get_lock(name)
         with lock1:
             self.assertFalse(lock2.acquire(blocking=False))
 
     def test_get_lock_context_fails(self):
-        name = self._get_random_uuid()
+        name = tests.get_random_uuid()
         lock1 = self._coord.get_lock(name)
         lock2 = self._coord.get_lock(name)
         with mock.patch.object(lock2, 'acquire', return_value=False):
@@ -798,34 +893,34 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
                     lock2.__enter__)
 
     def test_get_lock_context_check_value(self):
-        name = self._get_random_uuid()
+        name = tests.get_random_uuid()
         lock = self._coord.get_lock(name)
         with lock as returned_lock:
             self.assertEqual(lock, returned_lock)
 
     def test_lock_context_manager_acquire_no_argument(self):
-        name = self._get_random_uuid()
+        name = tests.get_random_uuid()
         lock1 = self._coord.get_lock(name)
         lock2 = self._coord.get_lock(name)
         with lock1():
             self.assertFalse(lock2.acquire(blocking=False))
 
     def test_lock_context_manager_acquire_argument_return_value(self):
-        name = self._get_random_uuid()
+        name = tests.get_random_uuid()
         blocking_value = 10.12
         lock = self._coord.get_lock(name)
         with lock(blocking_value) as returned_lock:
             self.assertEqual(lock, returned_lock)
 
     def test_lock_context_manager_acquire_argument_release_within(self):
-        name = self._get_random_uuid()
+        name = tests.get_random_uuid()
         blocking_value = 10.12
         lock = self._coord.get_lock(name)
         with lock(blocking_value) as returned_lock:
             self.assertTrue(returned_lock.release())
 
     def test_lock_context_manager_acquire_argument(self):
-        name = self._get_random_uuid()
+        name = tests.get_random_uuid()
         blocking_value = 10.12
         lock = self._coord.get_lock(name)
         with mock.patch.object(lock, 'acquire', wraps=True, autospec=True) as \
@@ -834,7 +929,7 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
                 mock_acquire.assert_called_once_with(blocking_value)
 
     def test_lock_context_manager_acquire_argument_timeout(self):
-        name = self._get_random_uuid()
+        name = tests.get_random_uuid()
         lock1 = self._coord.get_lock(name)
         lock2 = self._coord.get_lock(name)
         with lock1:
@@ -845,13 +940,13 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
                 pass
 
     def test_get_lock_locked_twice(self):
-        name = self._get_random_uuid()
+        name = tests.get_random_uuid()
         lock = self._coord.get_lock(name)
         with lock:
             self.assertFalse(lock.acquire(blocking=False))
 
     def test_get_multiple_locks_with_same_coord(self):
-        name = self._get_random_uuid()
+        name = tests.get_random_uuid()
         lock1 = self._coord.get_lock(name)
         lock2 = self._coord.get_lock(name)
         self.assertTrue(lock1.acquire())
@@ -860,19 +955,19 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         self.assertTrue(lock1.release())
 
     def test_ensure_acquire_release_return(self):
-        name = self._get_random_uuid()
+        name = tests.get_random_uuid()
         lock1 = self._coord.get_lock(name)
         self.assertTrue(lock1.acquire())
         self.assertTrue(lock1.release())
         self.assertFalse(lock1.release())
 
     def test_get_lock_multiple_coords(self):
-        member_id2 = self._get_random_uuid()
+        member_id2 = tests.get_random_uuid()
         client2 = tooz.coordination.get_coordinator(self.url,
                                                     member_id2)
         client2.start()
 
-        lock_name = self._get_random_uuid()
+        lock_name = tests.get_random_uuid()
         lock = self._coord.get_lock(lock_name)
         self.assertTrue(lock.acquire())
 
@@ -889,34 +984,18 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         self._coord.start()
 
     def do_test_name_property(self):
-        name = self._get_random_uuid()
+        name = tests.get_random_uuid()
         lock = self._coord.get_lock(name)
         self.assertEqual(name, lock.name)
 
-    @staticmethod
-    def _get_random_uuid():
-        return str(uuid.uuid4()).encode('ascii')
-
     def test_acquire_twice_no_deadlock_releasing(self):
-        name = self._get_random_uuid()
+        name = tests.get_random_uuid()
         lock = self._coord.get_lock(name)
         self.assertTrue(lock.acquire(blocking=False))
         self.assertFalse(lock.acquire(blocking=False))
         self.assertTrue(lock.release())
 
 
-class ZakeTestAPI(TestAPI):
-    url = "zake://"
-
-
-class IPCTestAPI(TestAPI):
-    url = "ipc://"
-
-
-class FileTestAPI(TestAPI):
-    url = "file:///tmp"
-
-
 class TestHook(testcase.TestCase):
     def setUp(self):
         super(TestHook, self).setUp()
