@@ -14,19 +14,15 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-import uuid
-
-try:
-    # Added in python 3.3+
-    from unittest import mock
-except ImportError:
-    import mock
+from unittest import mock
 
 from oslo_utils import encodeutils
 import testtools
 from testtools import testcase
 
+import tooz
 from tooz import coordination
+from tooz import tests
 
 # Handle the case gracefully where the driver is not installed.
 try:
@@ -48,7 +44,7 @@ class TestPostgreSQLFailures(testcase.TestCase):
         def _safe_stop(coord):
             try:
                 coord.stop()
-            except coordination.ToozError as e:
+            except tooz.ToozError as e:
                 # TODO(harlowja): make this better, so that we don't have to
                 # do string checking...
                 message = encodeutils.exception_to_unicode(e)
@@ -57,7 +53,7 @@ class TestPostgreSQLFailures(testcase.TestCase):
                     raise
 
         coord = coordination.get_coordinator(self.FAKE_URL,
-                                             str(uuid.uuid4()).encode('ascii'))
+                                             tests.get_random_uuid())
         self.addCleanup(_safe_stop, coord)
         return coord
 
@@ -89,7 +85,7 @@ class TestPostgreSQLFailures(testcase.TestCase):
         c = self._create_coordinator()
         c.start()
         test_lock = c.get_lock(b'test-lock')
-        self.assertRaises(coordination.ToozError, test_lock.acquire)
+        self.assertRaises(tooz.ToozError, test_lock.acquire)
 
     @mock.patch("tooz.drivers.pgsql.psycopg2.connect")
     def test_failure_release_lock(self, psycopg2_connector):
@@ -111,4 +107,4 @@ class TestPostgreSQLFailures(testcase.TestCase):
         c.start()
         test_lock = c.get_lock(b'test-lock')
         self.assertTrue(test_lock.acquire())
-        self.assertRaises(coordination.ToozError, test_lock.release)
+        self.assertRaises(tooz.ToozError, test_lock.release)
